@@ -1,11 +1,11 @@
 /*
  * textsearch_senna.c
  *
- *	Copyright (c) 2009, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ *	Copyright (c) 2009-2010, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 #include "postgres.h"
 
-#include <senna.h>
+#include <senna/senna.h>
 
 #include <fcntl.h>
 #include <sys/stat.h>
@@ -25,7 +25,15 @@
 #include "miscadmin.h"
 #include "nodes/parsenodes.h"
 #include "nodes/relation.h"
+#include "parser/parsetree.h"
+
+#if PG_VERSION_NUM >= 80300
 #include "postmaster/syslogger.h"
+#else
+/* 8.2 does not install postmaster headers !? */
+extern DLLIMPORT char *Log_directory;
+#endif
+
 #include "storage/fd.h"
 #include "storage/lmgr.h"
 #include "storage/bufmgr.h"
@@ -34,7 +42,11 @@
 #include "utils/inval.h"
 #include "utils/lsyscache.h"
 #include "utils/memutils.h"
+#include "utils/relcache.h"
 #include "utils/selfuncs.h"
+#include "utils/syscache.h"
+
+#include "pgut/pgut-be.h"
 
 PG_MODULE_MAGIC;
 
@@ -54,11 +66,14 @@ PG_MODULE_MAGIC;
 		message, \
 		ItemPointerGetBlockNumber(ctid), \
 		ItemPointerGetOffsetNumber(ctid))
+#define DEBUG_RECORDS_CACHE(found) \
+	elog(DEBUG_LEVEL, "senna: records is %scached", (found ? "" : "NOT "))
 #else
 #define DEBUG_INDEX_CREATE(rnode)		((void) 0)
 #define DEBUG_INDEX_OPEN(rnode)			((void) 0)
 #define DEBUG_INDEX_CLOSE(rnode)		((void) 0)
 #define DEBUG_DELETE(message, ctid)		((void) 0)
+#define DEBUG_RECORDS_CACHE(found)		((void) 0)
 #endif
 
 #define INDEX_CACHE_SIZE	4
@@ -71,6 +86,11 @@ PG_MODULE_MAGIC;
  */
 #define VACUUM_USES_HEAP
 
+/*
+ * RESTSEL_USES_SCAN - do actual scan during estimation.
+ */
+#define RESTSEL_USES_SCAN
+
 typedef struct SennaOptions
 {
 	int32       vl_len_;	/* varlena header (do not touch directly!) */
@@ -90,9 +110,17 @@ typedef struct SennaQueryCache
 	sen_query	   *query;
 } SennaQueryCache;
 
+typedef struct SennaRecordsCache
+{
+	sen_records	   *records;
+	sen_query	   *query;
+	Oid				relid;
+} SennaRecordsCache;
+
 typedef struct SennaScanDesc
 {
 	sen_records	   *records;
+	bool			recheck;
 #if NOT_USED
 	sen_sort_optarg	sorter;
 #endif
@@ -108,6 +136,7 @@ PG_FUNCTION_INFO_V1(senna_drop_index);
 PG_FUNCTION_INFO_V1(senna_reindex_index);
 PG_FUNCTION_INFO_V1(senna_contains);
 PG_FUNCTION_INFO_V1(senna_contained);
+PG_FUNCTION_INFO_V1(senna_restsel);
 PG_FUNCTION_INFO_V1(senna_insert);
 PG_FUNCTION_INFO_V1(senna_beginscan);
 PG_FUNCTION_INFO_V1(senna_gettuple);
@@ -126,89 +155,54 @@ PG_FUNCTION_INFO_V1(senna_costestimate);
 PG_FUNCTION_INFO_V1(senna_options);
 #endif
 
-extern void		_PG_init(void);
-extern void		_PG_fini(void);
-extern Datum pg_sync_file(PG_FUNCTION_ARGS);
-extern Datum senna_drop_index(PG_FUNCTION_ARGS);
-extern Datum senna_reindex_index(PG_FUNCTION_ARGS);
-extern Datum senna_contains(PG_FUNCTION_ARGS);
-extern Datum senna_contained(PG_FUNCTION_ARGS);
-extern Datum senna_insert(PG_FUNCTION_ARGS);
-extern Datum senna_beginscan(PG_FUNCTION_ARGS);
-extern Datum senna_gettuple(PG_FUNCTION_ARGS);
+extern void PGUT_EXPORT _PG_init(void);
+extern void PGUT_EXPORT _PG_fini(void);
+extern Datum PGUT_EXPORT pg_sync_file(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_drop_index(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_reindex_index(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_contains(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_contained(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_restsel(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_insert(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_beginscan(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_gettuple(PG_FUNCTION_ARGS);
 #if PG_VERSION_NUM >= 80400
-extern Datum senna_getbitmap(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_getbitmap(PG_FUNCTION_ARGS);
 #else
-extern Datum senna_getmulti(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_getmulti(PG_FUNCTION_ARGS);
 #endif
-extern Datum senna_rescan(PG_FUNCTION_ARGS);
-extern Datum senna_endscan(PG_FUNCTION_ARGS);
-extern Datum senna_build(PG_FUNCTION_ARGS);
-extern Datum senna_bulkdelete(PG_FUNCTION_ARGS);
-extern Datum senna_vacuumcleanup(PG_FUNCTION_ARGS);
-extern Datum senna_costestimate(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_rescan(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_endscan(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_build(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_bulkdelete(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_vacuumcleanup(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_costestimate(PG_FUNCTION_ARGS);
 #if PG_VERSION_NUM >= 80200
-extern Datum senna_options(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_options(PG_FUNCTION_ARGS);
 #endif
 
 static char *SennaPath(const RelFileNode *rnode);
 static sen_encoding SennaEncoding(void);
-static sen_index *SennaIndexOpen(Relation index, bool create);
+static sen_index *SennaIndexOpen(Relation index, int flags);
 static void SennaIndexClose(Oid relid, const RelFileNode *rnode);
 static bool SennaRemove(const RelFileNode *rnode);
 static sen_query *SennaQuery(const char *key, size_t len);
-static sen_records *SennaRecordsOpen(void);
+static sen_records *SennaRecordsOpen(Relation indexRelation, const char *str, int len, sen_index **idx);
 static void SennaRecordsClose(sen_records *records);
 static void SennaLock(Relation index, LOCKMODE lockmode);
 static void SennaUnlock(Relation index, LOCKMODE lockmode);
-static void SennaScanOpen(SennaScanDesc *desc, IndexScanDesc scan, Datum datum);
-static bool SennaScanNext(SennaScanDesc *desc, ItemPointer ctid);
+static bool SennaScanNext(SennaScanDesc *desc, IndexScanDesc scan, ItemPointer ctid);
 static void SennaScanClose(SennaScanDesc *desc);
 static void SennaInsert(sen_index *index, ItemPointer ctid, Datum value);
 static void SennaDelete(sen_index *index, ItemPointer key, text *value);
 static void SennaBuildCallback(Relation index, HeapTuple htup, Datum *values, bool *isnull, bool tupleIsAlive, void *context);
-static size_t SennaIndexSize(sen_index *index);
+static uint64 SennaIndexSize(sen_index *index);
 static IndexBulkDeleteResult *SennaBulkDeleteResult(sen_index *index);
 static void SennaXactCallback(XactEvent event, void *arg);
 static void SennaCacheRelCallback(Datum arg, Oid relid);
 
-#if PG_VERSION_NUM < 80300
-
-#define PG_GETARG_TEXT_PP(n)		PG_GETARG_TEXT_P(n)
-#define VARSIZE_ANY_EXHDR(v)		(VARSIZE(v) - VARHDRSZ)
-#define VARDATA_ANY(v)				VARDATA(v)
-#define SET_VARSIZE(v, sz)			(VARATT_SIZEP(v) = (sz))
-#define pg_detoast_datum_packed(v)	pg_detoast_datum(v)
-#define DatumGetTextPP(v)			DatumGetTextP(v)
-#define ItemIdIsNormal(v)			ItemIdIsUsed(v)
-#define IndexBuildHeapScan(heap, index, info, sync, callback, state) \
-	IndexBuildHeapScan((heap), (index), (info), (callback), (state))
-
-#endif
-
 #if PG_VERSION_NUM < 80400
 
-#define MAIN_FORKNUM	0
-#define relpath(rnode, forknum)		relpath((rnode))
-
-static char *
-text_to_cstring(const text *t)
-{
-	/* must cast away the const, unfortunately */
-	text	   *tunpacked = pg_detoast_datum_packed((struct varlena *) t);
-	int			len = VARSIZE_ANY_EXHDR(tunpacked);
-	char	   *result;
-
-	result = (char *) palloc(len + 1);
-	memcpy(result, VARDATA_ANY(tunpacked), len);
-	result[len] = '\0';
-
-	if (tunpacked != t)
-		pfree(tunpacked);
-	
-	return result;
-}
-
 static void
 RemoveRelations(DropStmt *drop)
 {
@@ -219,17 +213,28 @@ RemoveRelations(DropStmt *drop)
 
 #endif
 
+#define STRATEGY_IS_LIKE(s)	((s) == 3 || (s) == 4)
+
+#define OPCLASS_IS_NORM(opr) 	(((opr)[0] || (opr)[1]) && !((opr)[2]) && !((opr)[3]))
+#define OPCLASS_IS_LIKE(opr) 	(((opr)[2] || (opr)[3]) && !((opr)[0]) && !((opr)[1]))
+
+#define FLAGS_NORM		(SEN_INDEX_NGRAM | SEN_INDEX_NORMALIZE)
+#define FLAGS_LIKE		(SEN_INDEX_NGRAM | SEN_INDEX_SPLIT_ALPHA | SEN_INDEX_SPLIT_DIGIT | SEN_INDEX_SPLIT_SYMBOL)
+
 static SennaIndexCache	index_cache[INDEX_CACHE_SIZE];
 static SennaQueryCache	query_cache[QUERY_CACHE_SIZE];
-static List	*records_list;
+static List			   *records_cache;	/* List of SennaRecordsCache */
 
 #if PG_VERSION_NUM >= 80400
 static int RELOPT_KIND_SENNA;
 #endif
 
+#define DEFAULT_INITIAL_N_SEGMENTS			512
+
 #define SennaInitialSegments(relation) \
 	((relation)->rd_options ? \
-	 ((SennaOptions *) (relation)->rd_options)->initial_n_segments : 8)
+	 ((SennaOptions *) (relation)->rd_options)->initial_n_segments : \
+	 DEFAULT_INITIAL_N_SEGMENTS)
 
 /* copied from adt/genfile.c */
 static char *
@@ -302,6 +307,12 @@ pg_sync_file(PG_FUNCTION_ARGS)
 	PG_RETURN_VOID();
 }
 
+static bool
+is_senna_index(Relation index)
+{
+	return strcmp(NameStr(index->rd_am->amname), "senna") == 0;
+}
+
 /*
  * Open senna index, or raise error if the index is not a senna.
  */
@@ -309,7 +320,6 @@ static Relation
 senna_index_open(Oid relid, LOCKMODE lockmode)
 {
 	Relation	index;
-	const char *type;
 
 	index = index_open(relid, lockmode);
 
@@ -318,12 +328,11 @@ senna_index_open(Oid relid, LOCKMODE lockmode)
 		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_CLASS,
 					   RelationGetRelationName(index));
 
-	type = NameStr(index->rd_am->amname);
-
-	if (strcmp(type, "senna") != 0)
+	if (!is_senna_index(index))
 		ereport(ERROR,
 				(errcode(ERRCODE_WRONG_OBJECT_TYPE),
-				 errmsg("senna: not a senna index (%s)", type)));
+				 errmsg("senna: not a senna index (%s)",
+						NameStr(index->rd_am->amname))));
 
 	return index;
 }
@@ -381,7 +390,7 @@ senna_reindex_index(PG_FUNCTION_ARGS)
 	rnode = index->rd_node;
 	index_close(index, NoLock);
 
-	reindex_index(relid);
+	reindex_index(relid, false);
 
 	SennaRemove(&rnode);
 
@@ -392,7 +401,7 @@ static bool
 SennaContains(text *doc, text *query)
 {
 	const char *str = VARDATA_ANY(doc);
-	int			len = VARSIZE_ANY_EXHDR(doc);
+	unsigned	len = VARSIZE_ANY_EXHDR(doc);
 	sen_query  *q;
 	sen_rc		rc;
 	int			found;
@@ -421,6 +430,106 @@ senna_contained(PG_FUNCTION_ARGS)
 		PG_GETARG_TEXT_PP(1), PG_GETARG_TEXT_PP(0)));
 }
 
+/*
+ * selectivity for %% and @@ operator.
+ */
+Datum
+senna_restsel(PG_FUNCTION_ARGS)
+{
+#ifdef RESTSEL_USES_SCAN
+	PlannerInfo *root = (PlannerInfo *) PG_GETARG_POINTER(0);
+#ifdef NOT_USED
+	Oid			operator = PG_GETARG_OID(1);
+#endif
+	List	   *args = (List *) PG_GETARG_POINTER(2);
+	int			varRelid = PG_GETARG_INT32(3);
+	VariableStatData vardata;
+	Node	   *other;
+	bool		varonleft;
+	double		selec;
+
+	if (!get_restriction_variable(root, args, varRelid, &vardata, &other, &varonleft))
+	{
+		selec = DEFAULT_MATCH_SEL;
+	}
+	else if (!IsA(other, Const))
+	{
+		/* no idea if non-const... */
+		selec = 1.0 / get_variable_numdistinct(&vardata);
+	}
+	else if (((Const *) other)->constisnull)
+	{
+		/* variable = NULL will never return TRUE. */
+		selec = 0.0;
+	}
+	else if (vardata.rel == NULL || vardata.rel->rtekind != RTE_RELATION)
+	{
+		/* no idea if non-heap... */
+		selec = DEFAULT_MATCH_SEL;
+	}
+	else
+	{
+		Oid			relid;
+		Relation	heapRel;
+		ListCell   *cell;
+
+		selec = DEFAULT_MATCH_SEL;
+
+		/* Find a senna index for on the heap */
+		relid = planner_rt_fetch(vardata.rel->relid, root)->relid;
+		heapRel = heap_open(relid, AccessShareLock);
+		foreach(cell, RelationGetIndexList(heapRel))
+		{
+			Oid			indrelid = lfirst_oid(cell);
+			Relation	indexRel;
+
+			indexRel = index_open(indrelid, AccessShareLock);
+			if (is_senna_index(indexRel))
+			{
+				text		   *query;
+				sen_index	   *index;
+				sen_records	   *records;
+				unsigned int	nhits;
+				unsigned int	nkeys;
+
+				/*
+				 * If we find a senna index, we search values actually with
+				 * the index. We don't free the records here so that the result
+				 * records will be cached and we would reuse them in executor.
+				 * Cleanup will be performed at the end of transaction.
+				 */
+				query = DatumGetTextPP(((Const *) other)->constvalue);
+				records = SennaRecordsOpen(indexRel,
+					VARDATA_ANY(query), VARSIZE_ANY_EXHDR(query), &index);
+				nhits = sen_records_nhits(records);
+				nkeys = sen_sym_size(index->keys);
+#ifdef NOT_USED
+				SennaRecordsClose(records);
+#endif
+				if (nhits >= nkeys)
+					selec = 1.0;
+				else if (nkeys > 0)
+					selec = (double)nhits / nkeys;
+
+				index_close(indexRel, AccessShareLock);
+				break;
+			}
+			index_close(indexRel, AccessShareLock);
+		}
+		heap_close(heapRel, AccessShareLock);
+	}
+
+	ReleaseVariableStats(vardata);
+
+	PG_RETURN_FLOAT8((float8) selec);
+
+#else	/* RESTSEL_USES_SCAN */
+
+	PG_RETURN_FLOAT8(DEFAULT_MATCH_SEL);
+
+#endif	/* RESTSEL_USES_SCAN */
+}
+
 Datum
 senna_insert(PG_FUNCTION_ARGS)
 {
@@ -439,7 +548,7 @@ senna_insert(PG_FUNCTION_ARGS)
 	if (isnull[0])
 		PG_RETURN_BOOL(false);
 
-	i = SennaIndexOpen(index, false);
+	i = SennaIndexOpen(index, 0);
 
 	SennaLock(index, ExclusiveLock);
 	SennaInsert(i, ctid, values[0]);
@@ -453,7 +562,7 @@ senna_beginscan(PG_FUNCTION_ARGS)
 {
 	Relation		index = (Relation) PG_GETARG_POINTER(0);
 	int				keysz = PG_GETARG_INT32(1);
-	ScanKey			scankey = (ScanKey) PG_GETARG_POINTER(2);
+	ScanKey			key = (ScanKey) PG_GETARG_POINTER(2);
 	IndexScanDesc	scan;
 
 	if (keysz < 1)
@@ -461,7 +570,7 @@ senna_beginscan(PG_FUNCTION_ARGS)
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 				 errmsg("senna: do not support whole-index scans")));
 
-	scan = RelationGetIndexScan(index, keysz, scankey);
+	scan = RelationGetIndexScan(index, keysz, key);
 
 	PG_RETURN_POINTER(scan);
 }
@@ -514,7 +623,7 @@ senna_gettuple(PG_FUNCTION_ARGS)
 	IndexScanDesc	scan = (IndexScanDesc) PG_GETARG_POINTER(0);
 	ScanDirection	dir = (ScanDirection) PG_GETARG_INT32(1);
 
-	SennaScanDesc  *desc = (SennaScanDesc *) scan->opaque;
+	SennaScanDesc *desc = (SennaScanDesc *) scan->opaque;
 	ItemPointerData ctid;
 
 	if (dir != ForwardScanDirection)
@@ -522,7 +631,7 @@ senna_gettuple(PG_FUNCTION_ARGS)
 
 	if (scan->kill_prior_tuple)
 	{
-		sen_index  *index = SennaIndexOpen(scan->indexRelation, false);
+		sen_index  *index = SennaIndexOpen(scan->indexRelation, 0);
 		text	   *value = extract_value(&scan->xs_ctup,
 								scan->heapRelation, scan->indexRelation);
 
@@ -544,13 +653,14 @@ senna_gettuple(PG_FUNCTION_ARGS)
 	}
 #endif
 
-#if PG_VERSION_NUM >= 80400
-	scan->xs_recheck = false;
-#endif
-	if (!SennaScanNext(desc, &ctid))
+	if (!SennaScanNext(desc, scan, &ctid))
 		PG_RETURN_BOOL(false);
 
 	scan->xs_ctup.t_self = ctid;
+#if PG_VERSION_NUM >= 80400
+	scan->xs_recheck = desc->recheck;
+#endif
+
 	PG_RETURN_BOOL(true);
 }
 
@@ -562,14 +672,14 @@ senna_getbitmap(PG_FUNCTION_ARGS)
 	IndexScanDesc	scan = (IndexScanDesc) PG_GETARG_POINTER(0);
 	TIDBitmap	   *tbm = (TIDBitmap *) PG_GETARG_POINTER(1);
 
-	SennaScanDesc  *desc = (SennaScanDesc *) scan->opaque;
+	SennaScanDesc *desc = (SennaScanDesc *) scan->opaque;
 	int64			ntids;
 	ItemPointerData ctid;
 
-	for (ntids = 0; SennaScanNext(desc, &ctid); ntids++)
+	for (ntids = 0; SennaScanNext(desc, scan, &ctid); ntids++)
 	{
 		CHECK_FOR_INTERRUPTS();
-		tbm_add_tuples(tbm, &ctid, 1, false);
+		tbm_add_tuples(tbm, &ctid, 1, desc->recheck);
 	}
 
 	PG_RETURN_INT64(ntids);
@@ -585,32 +695,15 @@ senna_getmulti(PG_FUNCTION_ARGS)
 	int32		max_tids = PG_GETARG_INT32(2);
 	int32	   *returned_tids = (int32 *) PG_GETARG_POINTER(3);
 
-	SennaScanDesc  *desc = (SennaScanDesc *) scan->opaque;
+	SennaScanDesc *desc = (SennaScanDesc *) scan->opaque;
 	int64			ntids;
 	ItemPointerData ctid;
 
 	if (max_tids <= 0)			/* behave correctly in boundary case */
 		PG_RETURN_BOOL(true);
 
-#if 0
-	/* If we haven't started the scan yet, fetch the first page & tuple. */
-	if (!BTScanPosIsValid(so->currPos))
-	{
-		res = _bt_first(scan, ForwardScanDirection);
-		if (!res)
-		{
-			/* empty scan */
-			*returned_tids = ntids;
-			PG_RETURN_BOOL(res);
-		}
-		/* Save tuple ID, and continue scanning */
-		tids[ntids] = scan->xs_ctup.t_self;
-		ntids++;
-	}
-#endif
-
 	for (ntids = 0;
-		 ntids < max_tids && SennaScanNext(desc, &ctid);
+		 ntids < max_tids && SennaScanNext(desc, scan, &ctid);
 		 ntids++)
 	{
 		/* Save tuple ID, and continue scanning */
@@ -627,9 +720,9 @@ Datum
 senna_rescan(PG_FUNCTION_ARGS)
 {
 	IndexScanDesc	scan = (IndexScanDesc) PG_GETARG_POINTER(0);
-	ScanKey			scankey = (ScanKey) PG_GETARG_POINTER(1);
+	ScanKey			key = (ScanKey) PG_GETARG_POINTER(1);
 
-	SennaScanDesc  *desc = (SennaScanDesc  *) scan->opaque;
+	SennaScanDesc *desc = (SennaScanDesc *) scan->opaque;
 
 	if (desc == NULL)
 	{
@@ -642,8 +735,8 @@ senna_rescan(PG_FUNCTION_ARGS)
 	else
 		SennaScanClose(desc);
 
-	if (scankey && scan->numberOfKeys > 0)
-		SennaScanOpen(desc, scan, scankey[0].sk_argument);
+	if (key && scan->numberOfKeys > 0)
+		memmove(scan->keyData, key, scan->numberOfKeys * sizeof(ScanKeyData));
 
 	PG_RETURN_VOID();
 }
@@ -653,7 +746,7 @@ senna_endscan(PG_FUNCTION_ARGS)
 {
 	IndexScanDesc	scan = (IndexScanDesc) PG_GETARG_POINTER(0);
 
-	SennaScanDesc  *desc = (SennaScanDesc *) scan->opaque;
+	SennaScanDesc *desc = (SennaScanDesc *) scan->opaque;
 
 	if (desc != NULL)
 	{
@@ -671,6 +764,7 @@ senna_build(PG_FUNCTION_ARGS)
 	Relation	index = (Relation) PG_GETARG_POINTER(1);
 	IndexInfo  *indexInfo = (IndexInfo *) PG_GETARG_POINTER(2);
 	TupleDesc	tupdesc;
+	int			flags;
 
 	IndexBuildResult   *result;
 	SennaBuildState		state;
@@ -685,6 +779,18 @@ senna_build(PG_FUNCTION_ARGS)
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 				 errmsg("senna: do not support multi columns")));
 
+	if (OPCLASS_IS_NORM(index->rd_operator))
+		flags = FLAGS_NORM;
+	else if (OPCLASS_IS_LIKE(index->rd_operator))
+		flags = FLAGS_LIKE;
+	else
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("senna: unsupported operator class")));
+		flags = 0;	/* keep compilar quiet */
+	}
+
 	tupdesc = RelationGetDescr(index);
 	switch (tupdesc->attrs[0]->atttypid)
 	{
@@ -698,7 +804,7 @@ senna_build(PG_FUNCTION_ARGS)
 				 errmsg("senna: only supports char, varchar and text")));
 	}
 
-	state.index = SennaIndexOpen(index, true);
+	state.index = SennaIndexOpen(index, flags);
 
 	PG_TRY();
 	{
@@ -736,7 +842,7 @@ senna_bulkdelete(PG_FUNCTION_ARGS)
 	HeapTupleData	tuple;
 #endif
 
-	index = SennaIndexOpen(info->index, false);
+	index = SennaIndexOpen(info->index, 0);
 
 	if (stats == NULL)
 		stats = SennaBulkDeleteResult(index);
@@ -827,7 +933,7 @@ senna_vacuumcleanup(PG_FUNCTION_ARGS)
 	IndexBulkDeleteResult *stats = (IndexBulkDeleteResult *) PG_GETARG_POINTER(1);
 
 	if (stats == NULL)
-		stats = SennaBulkDeleteResult(SennaIndexOpen(info->index, false));
+		stats = SennaBulkDeleteResult(SennaIndexOpen(info->index, 0));
 
 	PG_RETURN_POINTER(stats);
 }
@@ -838,6 +944,8 @@ senna_costestimate(PG_FUNCTION_ARGS)
 	/*
 	 * We cannot use genericcostestimate because it is a static funciton.
 	 * Use gincostestimate instead, which just calls genericcostestimate.
+	 * Actual estimation is done in senna_restsel, which is selectivity
+	 * functions for %% and @@ operator.
 	 */
 	return gincostestimate(fcinfo);
 }
@@ -868,7 +976,6 @@ senna_options(PG_FUNCTION_ARGS)
 					validate, tab, lengthof(tab));
 
 	pfree(options);
-
 	PG_RETURN_BYTEA_P(rdopts);
 }
 
@@ -952,7 +1059,7 @@ SennaBuildCallback(Relation index,
 }
 
 static sen_index *
-SennaIndexOpen(Relation indexRelation, bool create)
+SennaIndexOpen(Relation indexRelation, int flags)
 {
 	int					i;
 	SennaIndexCache    *cache;
@@ -972,13 +1079,13 @@ SennaIndexOpen(Relation indexRelation, bool create)
 
 	/* Be careful not to leak resources on error. */
 	path = SennaPath(&indexRelation->rd_node);
-	if (create)
+	if (flags != 0)
 	{
 		DEBUG_INDEX_CREATE(indexRelation->rd_node);
 		index = sen_index_create(
 			path,
 			SizeOfIptrData,
-			SEN_INDEX_NGRAM | SEN_INDEX_NORMALIZE,
+			flags,
 			SennaInitialSegments(indexRelation),
 			SennaEncoding());
 		if (index == NULL)
@@ -1131,32 +1238,91 @@ SennaQuery(const char *sql, size_t len)
 }
 
 static sen_records *
-SennaRecordsOpen(void)
+SennaRecordsOpen(Relation indexRelation, const char *str, int len, sen_index **idx)
 {
-	MemoryContext	oldctx;
-	sen_records	   *records;
+	MemoryContext		oldctx;
+	sen_query		   *query;
+	sen_records		   *records;
+	sen_index		   *index;
+	sen_rc				rc;
+	ListCell		   *cell;
+	SennaRecordsCache	*cache;
+
+	query = SennaQuery(str, len);
+
+	/* search records cache first */
+	foreach(cell, records_cache)
+	{
+		cache = (SennaRecordsCache *) lfirst(cell);
+		if (cache->query == query &&
+			cache->relid == RelationGetRelid(indexRelation))
+		{
+			DEBUG_RECORDS_CACHE(true);
+			records = cache->records;
+			if (idx)
+				*idx = SennaIndexOpen(indexRelation, 0);
+			return records;
+		}
+	}
+
+	DEBUG_RECORDS_CACHE(false);
 
 	records = sen_records_open(sen_rec_document, sen_rec_none, 0);
 	if (records == NULL)
 		elog(ERROR, "senna: sen_records_open()");
 
 	oldctx = MemoryContextSwitchTo(TopMemoryContext);
-	records_list = lappend(records_list, records);
+	cache = palloc(sizeof(SennaRecordsCache));
+	cache->records = records;
+	cache->query = query;
+	cache->relid = RelationGetRelid(indexRelation);
+	records_cache = lappend(records_cache, cache);
 	MemoryContextSwitchTo(oldctx);
 
+	index = SennaIndexOpen(indexRelation, 0);
+
+	/*
+	 * Use sen_query_exec() instead of sen_index_select() in order to
+	 * minimize locking duration. Compile queries before locking.
+	 */
+	SennaLock(indexRelation, ShareLock);
+	rc = sen_query_exec(index, query, records, sen_sel_or);
+	SennaUnlock(indexRelation, ShareLock);
+
+	if (rc != sen_success)
+		elog(ERROR, "senna: sen_query_exec() : code=%d", rc);
+
+	if (idx)
+		*idx = index;
 	return records;
 }
 
+/*
+ * Release records from cache.
+ */
 static void
 SennaRecordsClose(sen_records *records)
 {
-	sen_rc rc;
+	ListCell   *cell;
+	ListCell   *prev;
+	sen_rc		rc;
 
-	rc = sen_records_close(records);
-	if (rc != sen_success)
-		elog(WARNING, "senna: sen_records_close() : code=%d", rc);
+	prev = NULL;
+	foreach(cell, records_cache)
+	{
+		SennaRecordsCache *cache = (SennaRecordsCache *) lfirst(cell);
+		if (cache->records == records)
+		{
+			rc = sen_records_close(cache->records);
+			if (rc != sen_success)
+				elog(WARNING, "senna: sen_records_close() : code=%d", rc);
+			records_cache = list_delete_cell(records_cache, cell, prev);
+			pfree(cache);
+			break;
+		}
 
-	records_list = list_delete_ptr(records_list, records);
+		prev = cell;
+	}
 }
 
 static void
@@ -1179,39 +1345,124 @@ SennaUnlock(Relation index, LOCKMODE lockmode)
 						 lockmode);
 }
 
-static void
-SennaScanOpen(SennaScanDesc *desc, IndexScanDesc scan, Datum datum)
+#define IS_LIKE_WILDCARD(c)		((c) == '%' || (c) == '_')
+#define LIKE_PRAGMA				"*D+ "
+#define LIKE_PRAGMA_LEN			4
+
+/*
+ * is_simple_like - true iff str matches %foo%.
+ *
+ * The foo part should not have '%', '_' nor any spaces.
+ */
+static bool
+is_simple_like(const char *str, int len)
 {
-	text		   *key;
-	sen_index	   *index;
-	sen_query	   *query;
-	sen_records	   *records = NULL;
-	sen_rc			rc;
+	int		i;
 
-	key = DatumGetTextPP(datum);
-	query = SennaQuery(VARDATA_ANY(key), VARSIZE_ANY_EXHDR(key));
+	if (len < 2 || str[0] != '%' || str[len - 1] != '%')
+		return false;
 
-	records = SennaRecordsOpen();
-
-	index = SennaIndexOpen(scan->indexRelation, false);
-
-	/*
-	 * Use sen_query_exec() instead of sen_index_select() in order to
-	 * minimize locking duration. Compile queries before locking.
-	 */
-	SennaLock(scan->indexRelation, ShareLock);
-	rc = sen_query_exec(index, query, records, sen_sel_or);
-	SennaUnlock(scan->indexRelation, ShareLock);
-
-	if (rc != sen_success)
-		elog(ERROR, "senna: sen_query_exec() : code=%d", rc);
-
-	desc->records = records;
+	for (i = 1; i < len - 1; i++)
+	{
+		if (IS_LIKE_WILDCARD(str[i]) ||
+			isspace((unsigned char) str[i]) ||
+			str[i] == '\\')
+			return false;
+	}
+	return true;
 }
 
-static bool
-SennaScanNext(SennaScanDesc *desc, ItemPointer ctid)
+static void
+SennaScanOpen(SennaScanDesc *desc, IndexScanDesc scan, StrategyNumber strategy, Datum datum)
 {
+	text		   *query;
+	const char	   *str;
+	int				len;
+
+	query = DatumGetTextPP(datum);
+	str = VARDATA_ANY(query);
+	len = VARSIZE_ANY_EXHDR(query);
+
+	/* queyr is LIKE. convert LIKE query to senna query */
+	if (STRATEGY_IS_LIKE(strategy))
+	{
+		if (is_simple_like(str, len))
+		{
+			/* don't need recheck if simple %foo% query */
+			str += 1;
+			len -= 2;
+		}
+		else
+		{
+			char	   *buffer;
+			char	   *s;
+			int			i;
+			int			wilds;
+
+			/* LIKE query needs recheck */
+			desc->recheck = true;
+
+			buffer = malloc(len + LIKE_PRAGMA_LEN);
+			memcpy(buffer, LIKE_PRAGMA, LIKE_PRAGMA_LEN);
+
+			/* skip wildcard at head */
+			while (IS_LIKE_WILDCARD(*str))
+			{
+				str++;
+				len--;
+			}
+
+			/* copy query. wildcards are converted to '+' */
+			for (i = 0, s = buffer + LIKE_PRAGMA_LEN, wilds = 0; i < len; i++)
+			{
+				if (str[i] == '\\')
+				{
+					if (i >= len - 1)
+						ereport(ERROR,
+							(errcode(ERRCODE_INVALID_ESCAPE_SEQUENCE),
+							 errmsg("LIKE pattern must not end with escape character")));
+					/* skip the escape sequence and copy the next char. */
+					*s++ = str[++i];
+					wilds = 0;
+				}
+				else if (IS_LIKE_WILDCARD(str[i]))
+				{
+					if (wilds == 0)
+						*s++ = ' ';
+					wilds++;
+				}
+				else
+				{
+					*s++ = str[i];
+					wilds = 0;
+				}
+			}
+			str = buffer;
+			len = s - buffer - wilds;
+		}
+	}
+
+	desc->records = SennaRecordsOpen(scan->indexRelation, str, len, NULL);
+}
+
+/*
+ * Scan senna index or return the next tuple.
+ */
+static bool
+SennaScanNext(SennaScanDesc *desc, IndexScanDesc scan, ItemPointer ctid)
+{
+	if (desc->records == NULL)
+	{
+		/* NULL key is not supported */
+		if (scan->keyData[0].sk_flags & SK_ISNULL)
+			return false;
+
+		SennaScanOpen(desc, scan,
+					  scan->keyData[0].sk_strategy,
+					  scan->keyData[0].sk_argument);
+	}
+
+	Assert(desc->records != NULL);
 	return sen_records_next(desc->records, ctid, SizeOfIptrData, NULL) == SizeOfIptrData;
 }
 
@@ -1275,7 +1526,7 @@ SennaDelete(sen_index *index, ItemPointer ctid, text *value)
 /*
  * Total index size in bytes.
  */
-static size_t
+static uint64
 SennaIndexSize(sen_index *index)
 {
 	unsigned file_size_keys;
@@ -1324,6 +1575,22 @@ SennaXactCallback(XactEvent event, void *arg)
 	sen_rc		rc;
 	int			i;
 
+	/* cleanup records */
+	if (records_cache != NIL)
+	{
+		ListCell   *cell;
+
+		foreach(cell, records_cache)
+		{
+			SennaRecordsCache *cache = (SennaRecordsCache *) lfirst(cell);
+			rc = sen_records_close(cache->records);
+			if (rc != sen_success)
+				elog(WARNING, "senna: sen_records_close() : code=%d", rc);
+		}
+		list_free_deep(records_cache);
+		records_cache = NIL;
+	}
+
 	/* cleanup queries */
 	for (i = 0; i < lengthof(query_cache); i++)
 	{
@@ -1342,21 +1609,6 @@ SennaXactCallback(XactEvent event, void *arg)
 			cache->key = NULL;
 		}
 	}
-
-	/* cleanup records */
-	if (records_list != NIL)
-	{
-		ListCell   *cell;
-
-		foreach(cell, records_list)
-		{
-			rc = sen_records_close((sen_records *) lfirst(cell));
-			if (rc != sen_success)
-				elog(WARNING, "senna: sen_records_close() : code=%d", rc);
-		}
-		list_free(records_list);
-		records_list = NIL;
-	}
 }
 
 static void
@@ -1407,6 +1659,11 @@ _PG_init(void)
 
 #if PG_VERSION_NUM >= 80400
 	RELOPT_KIND_SENNA = add_reloption_kind();
+	add_int_reloption(RELOPT_KIND_SENNA,
+		"initial_n_segments",
+		"initial size of .sen.i",
+		DEFAULT_INITIAL_N_SEGMENTS,
+		1, INT_MAX);
 #endif
 
 	sen_logger_info_set(&senna_logger);
